#include <math.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>
#include <sched.h>
#include <sys/resource.h>
#include <stdarg.h>

#include "940/940.h"

extern int g_hMemory;
extern unsigned short *g_pusRegs;

static pthread_t g_SoundThread;
static volatile int g_bSoundThreadRunning = DZZFALSE;
static volatile int g_bExitSoundThread = DZZFALSE;
static unsigned char *g_pSharedMemory = 0;
static SharedVariables *g_pShared;

void Pause940(int n)
{
	if(n)
		g_pusRegs[0x0904>>1] &= 0xFFFE;
	else
		g_pusRegs[0x0904>>1] |= 1;
} 

void Reset940(int yes)
{
	g_pusRegs[0x3B48>>1] = ((yes&1) << 7) | (0x03);
}

void Startup940()
{
	int nLen, nRead;
	FILE *fp;
	unsigned char ucData[1000];

	Reset940(1);
	Pause940(1);
  g_pusRegs[0x3B40>>1] = 0;
  g_pusRegs[0x3B42>>1] = 0;
  g_pusRegs[0x3B44>>1] = 0xffff;
  g_pusRegs[0x3B46>>1] = 0xffff;

	// load code940.bin
	nLen = 0;
	fp = fopen("code940.bin", "r");
	if(!fp)
		return;
	while(1)
	{
		nRead = fread(ucData, 1, 1000, fp);
		if(nRead <= 0)
			break;
		memcpy(g_pSharedMemory + nLen, ucData, nRead);
		nLen += nRead;
	}
	fclose(fp);
	
	Reset940(0);
	Pause940(0);

	usleep(10000);
}

void Shutdown940()
{
	Reset940(1);
	Pause940(1);
}

static void *SoundThread(void *p)
{
	int fd;
	int nRate, nBits, nStereo, nFrag;
	audio_buf_info abi;
	int nCurBuf = 0;

	g_bSoundThreadRunning = DZZTRUE;

	fd = open("/dev/dsp", O_WRONLY);
	if(fd <= 0)
	{
		g_bSoundThreadRunning = DZZFALSE;
		return NULL;
	}
	nRate = 44100;
	nBits = 16;
	nStereo = 1;
	nFrag = 0x2000B;

	ioctl(fd, SNDCTL_DSP_SPEED, &nRate);
	ioctl(fd, SNDCTL_DSP_SETFMT, &nBits);
	ioctl(fd, SNDCTL_DSP_STEREO, &nStereo);
	ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &nFrag);
	ioctl(fd, SNDCTL_DSP_GETOSPACE, &abi);

	// initialize the global variables

	g_pShared->g_bQuit = DZZFALSE;
	g_pShared->g_bMusicOn = DZZFALSE;
	g_pShared->g_bResetMusic = DZZFALSE;
	g_pShared->g_bMusicLoaded = DZZFALSE;

	g_pShared->g_pMusicLoopInPointer = (unsigned char *) 0x700000;
	g_pShared->g_nMusicLoopInLength = 0; // until it is loaded anyway

	// sound buffers
	g_pShared->g_bSoundBuffer0Full = DZZFALSE;
	g_pShared->g_bSoundBuffer1Full = DZZFALSE;

	memset(g_pShared->g_sSoundBuffer0, 0, sizeof(g_pShared->g_sSoundBuffer0));
	memset(g_pShared->g_sSoundBuffer1, 0, sizeof(g_pShared->g_sSoundBuffer1));

	Startup940();
	
	while(!g_bExitSoundThread)
	{
		if(((nCurBuf == 0) && g_pShared->g_bSoundBuffer0Full) ||
		   ((nCurBuf == 1) && g_pShared->g_bSoundBuffer1Full))
		{
			if((nCurBuf == 0) && g_pShared->g_bSoundBuffer0Full)
			{
				write(fd, g_pShared->g_sSoundBuffer0, 8192);
				nCurBuf = 1;
				g_pShared->g_bSoundBuffer0Full = DZZFALSE;
			}
			else if((nCurBuf == 1) && g_pShared->g_bSoundBuffer1Full)
			{
				write(fd, g_pShared->g_sSoundBuffer1, 8192);
				nCurBuf = 0;
				g_pShared->g_bSoundBuffer1Full = DZZFALSE;
			}
		}
	}
	close(fd);
	Shutdown940();
	g_bSoundThreadRunning = DZZFALSE;
	return NULL;
}

void StartMusic()
{
	g_pShared->g_bMusicOn = DZZTRUE;
}

void StopMusic()
{
	g_pShared->g_bMusicOn = DZZFALSE;
}

void ResetMusic()
{
	g_pShared->g_bResetMusic = DZZTRUE;
}

void LoadSong(char *pszFile)
{
	FILE *fp;
	int nLen;
	int n;
	unsigned char uc[100000];

	nLen = 0;
	fp = fopen(pszFile, "r");
	if(!fp)
		return;
	g_pShared->g_bMusicLoaded = DZZFALSE;
	while(1)
	{
		n = fread(uc, 1, 100000, fp);
		if(n <= 0)
			break;
		memcpy(g_pSharedMemory + (int) g_pShared->g_pMusicLoopInPointer + nLen, uc, n);
		nLen += n;
	}
	fclose(fp);
	g_pShared->g_nMusicLoopInLength = nLen;
	g_pShared->g_bResetMusic = DZZTRUE;
	g_pShared->g_bMusicLoaded = DZZTRUE;
}

void InitSound()
{
	Reset940(1);
	Pause940(1);

  g_pSharedMemory = (unsigned char *) mmap(0, 0xF80000, 
			PROT_READ|PROT_WRITE, MAP_SHARED, g_hMemory, 0x3000000);

	// the shared memory variable area lives at g_pSharedMemory + 0x200000
	g_pShared = (SharedVariables *) (g_pSharedMemory + 0x200000);

	pthread_create(&g_SoundThread, NULL, SoundThread, NULL);
	usleep(100000);
}

void CleanupSound()
{
	g_bExitSoundThread = DZZTRUE;
	while(g_bSoundThreadRunning)
		usleep(10000);
	if(g_pSharedMemory)
		munmap(g_pSharedMemory, 0xF80000);
	g_pSharedMemory  = 0;
}
